import functools

def adageop(func):
    """
    Decorator that adds a 's' attribute to a function
    The attribute can be used to partially define
    the function call, except for the 'adageobj'
    keyword argument, the return value is a
    single-argument ('adageobj') function
    """
    def partial(*args,**kwargs):
        return functools.partial(func,*args,**kwargs)
    func.s = partial
    return func

class Rule(object):
    def __init__(self,predicate,body):
        self.predicate = predicate
        self.body = body

    def applicable(self,adageobj):
        return self.predicate(adageobj)

    def apply(self,adageobj):
        return self.body(adageobj = adageobj)

def adagetask(func):
    """
    Decorator that adds a 's' attribute to a function
    The attribute can be used to fully define a function
    call to be executed at a later time. The result
    will be a zero-argument callable
    """
    try:
        from celery import shared_task
        func.celery = shared_task(func)
    except ImportError:
        pass

    def partial(*args,**kwargs):
        return functools.partial(func,*args,**kwargs)
    func.s = partial
    return func

def callbackrule(after = None):
    """
    A decorator that creates a adage Rule from a callback function
    The after argument is expected to container a dictionary
    of node identifiers. The callback is expected have two arguments
    A dictionary with the same keys as in after as keys, and the
    corresponding nodes as values, as well as the adajeobj will be
    passed to the callback
    """
    after = after or {}
    def decorator(func):
        def predicate(adageobj):
            return all([adageobj.dag.getNode(node).successful() for node in after.values()])
        def body(adageobj):
            depnodes = {k:adageobj.dag.getNode(v) for k,v in after.iteritems()}
            func(depnodes = depnodes, adageobj = adageobj)

        return Rule(predicate,body)
    return decorator



import logging
import time
import os
import shutil
import json
import subprocess
import dagstate
import nodestate
from datetime  import datetime
import networkx as nx
import adage.visualize as viz
import adage.serialize as serialize

class JSONDumpTracker(object):
    def __init__(self,dumpname, serializer = serialize.DefaultAdageEncoder):
        self.serializer = serializer
        self.dumpname = dumpname

    def initialize(self,adageobj):
        pass
        
    def track(self,adageobj):
        pass
    
    def finalize(self,adageobj):
        with open(self.dumpname,'w') as dumpfile:
            json.dump(adageobj,dumpfile, cls = self.serializer)

class GifTracker(object):
    def __init__(self,gifname,workdir,frames = 40):
        self.gifname = gifname
        self.workdir = workdir
        self.frames = frames
        
    def initialize(self,adageobj):
        pass
        
    def track(self,adageobj):
        pass
        
    def finalize(self,adageobj):
        if os.path.exists(self.workdir):
            shutil.rmtree(self.workdir)
        os.makedirs(self.workdir)
        for i in range(self.frames+1):
            viz.print_dag(adageobj.dag,'dag_{:02}'.format(i),self.workdir,time = i/float(self.frames))
        subprocess.check_call('convert -delay 50 $(ls {}/*.png|sort) {}'.format(self.workdir,self.gifname),shell = True)
        shutil.rmtree(self.workdir)
                
class TextSnapShotTracker(object):
    def __init__(self,logfilename,mindelta):
        self.logfilename = logfilename
        self.mindelta = mindelta
        self.last_update = None

    def initialize(self,adageobj):
        if not os.path.exists(os.path.dirname(self.logfilename)):
            os.makedirs(os.path.dirname(self.logfilename))
        with open(self.logfilename,'w') as logfile:
            timenow = datetime.now().isoformat()
            logfile.write('========== ADAGE LOG BEGIN at {} ==========\n'.format(timenow))
        self.update(adageobj)

    def track(self,adageobj):
        now = time.time()
        if not self.last_update or (now-self.last_update) > self.mindelta:
            self.last_update = now
            self.update(adageobj)

    def update(self,adageobj):
        dag = adageobj.dag
        with open(self.logfilename,'a') as logfile:
            logfile.write('---------- snapshot at {}\n'.format(datetime.now().isoformat()))
            for node in nx.topological_sort(dag):
                nodeobj = dag.getNode(node)
                submitted = nodeobj.submit_time is not None
                logfile.write('name: {} obj: {} submitted: {}\n'.format(
                        nodeobj.name,
                        nodeobj,
                        submitted
                    )
                )
    def finalize(self,adageobj):
        with open(self.logfilename,'a') as logfile:
            self.update(adageobj)
            timenow = datetime.now().isoformat()
            logfile.write('========== ADAGE LOG END at {} ==========\n'.format(timenow))

        
class SimpleReportTracker(object):
    def __init__(self, loggername, mindelta):
        self.log = logging.getLogger(loggername)
        self.mindelta = mindelta
        self.last_update = None
        
    def initialize(self,adageobj):
        pass

    def track(self,adageobj):
        now = time.time()
        if not self.last_update or (now-self.last_update) > self.mindelta:
            self.last_update = now
            self.update(adageobj)
    
    def finalize(self,adageobj):
        self.update(adageobj)

    def update(self,adageobj):
        dag, rules, applied = adageobj.dag, adageobj.rules, adageobj.applied_rules
        successful, failed, running, unsubmittable = 0, 0, 0, 0
        for node in dag.nodes():
            nodeobj = dag.getNode(node)
            if nodeobj.state == nodestate.RUNNING:
                running += 1
            if dagstate.node_status(nodeobj):
                successful+=1
            if dagstate.node_ran_and_failed(nodeobj):
                failed+=1
                self.log.error("node: {} failed. reason: {}".format(nodeobj,nodeobj.backend.fail_info(nodeobj.resultproxy)))
            if dagstate.upstream_failure(dag,nodeobj):
                unsubmittable+=1
        self.log.info('unsubmittable: {unsubmittable} | submitted: {submitted} | successful: {successful} | failed: {failed} | total: {total} | open rules: {rules} | applied rules: {applied}'.format(
            successful = successful,
            failed = failed,
            submitted = running,
            unsubmittable = unsubmittable,
            total =  len(dag.nodes()),
            rules = len(rules),
            applied = len(applied)))
        



import os
import time
import logging
import trackers

from adage.decorators import adageop, adagetask, Rule
from adage.adageobject import adageobject
from adage.pollingexec import setup_polling_execution
from adage.wflowcontroller import InMemoryController

#silence pyflakes
assert adageop
assert adagetask
assert Rule
assert adageobject

log = logging.getLogger(__name__)

def trackprogress(trackerlist,adageobj, method = 'track'):
    '''
    track adage workflow state using a list of trackers

    :param trackerlist: the stackers which should inspect the workflow
    :param adageobj: the adage workflow object
    :param method: tracking method to call. Must be one of ``initialize``, ``track``, ``finalize``
    :return: None
    '''
    map(lambda t: getattr(t,method)(adageobj), trackerlist)

def run_polling_workflow(controller, coroutine, update_interval, trackerlist = None, maxsteps = None):
    '''
    run the polling-style workflow by periodically checkinng if a graph can be extended or any noted be submitted
    runs validation 

    :param controller: the workflow controller
    :param coroutine: the adage coroutine to step through the workflow
    :param update_interval: time interval between workflow ticks
    :param trackerlist
    
    :return: None    
    :raises RuntimeError: if graph is finished and graph validation failed or workflow is failed (nodes unsuccessful)
    '''

    coroutine.send(controller)
    log.info('starting state loop.')

    try:
        trackprogress(trackerlist, controller.adageobj, method = 'initialize')
        for stepnum, controller in enumerate(coroutine):
            trackprogress(trackerlist, controller.adageobj)
            if maxsteps and (stepnum == maxsteps):
                log.info('reached number of maximum iterations ({})'.format(maxsteps))
                return
            time.sleep(update_interval)
    except:
        log.exception('some weird exception caught in adage process loop')
        raise
    finally:
        trackprogress(trackerlist, controller.adageobj, method = 'finalize')

    log.info('adage state loop done.')

    if not controller.validate():
        raise RuntimeError('DAG execution not validating')
    log.info('execution valid. (in terms of execution order)')

    if not controller.successful():
        log.error('raising RunTimeError due to failed jobs')
        raise RuntimeError('DAG execution failed')


    log.info('workflow completed successfully.')

def default_trackerlist(gif_workdir = None, text_loggername = __name__, texttrack_delta = 1):
    workdir = gif_workdir or os.getcwd()
    trackerlist  = [trackers.SimpleReportTracker(text_loggername, texttrack_delta)]
    trackerlist += [trackers.GifTracker(gifname = '{}/workflow.gif'.format(workdir), workdir = '{}/track'.format(workdir))]
    trackerlist += [trackers.TextSnapShotTracker(logfilename = '{}/adagesnap.txt'.format(workdir), mindelta = texttrack_delta)]
    return trackerlist

def rundag(adageobj = None,
           backend = None,
           extend_decider = None,
           submit_decider = None,
           update_interval = 0.01,
           loggername = __name__,
           trackevery = 1,
           workdir = None,
           default_trackers = True,
           additional_trackers = None,
           controller = None,
           maxsteps = None):
    '''
    Main adage entrypoint. It's a convenience wrapper around the main adage coroutine loop and
    sets up the backend, logging, tracking (for GIFs, Text Snapshots, etc..) and possible interactive
    hooks into the coroutine

    :param adageobj: the adage workflow object
    :param backend: the task execution backend to which to submit node tasks
    :param extend_decider: decision coroutine to deal with whether to extend the workflow graph
    :param submit_decider: decision coroutine to deal with whether to submit node tasks
    :param update_interval: minimum looping interval for main adage loop 
    :param loggername: python logger to use
    :param trackevery: tracking interval for default simple report tracker
    :param workdir: workdir for default visual tracker
    :param default_trackers: whether to enable default trackers (simple report, gif visualization, text snapshot)
    :param additional_trackers: list of any additional tracking objects
    :param controller: optional external controller (instead of adageobj parameter)
    :param maxsteps: maximum number of steps in polling-style workflow
    '''
    if loggername:
        global log
        log = logging.getLogger(loggername)

    ## get primed coroutine for polling-style workflow execution
    coroutine = setup_polling_execution(extend_decider, submit_decider)

    ## prepare tracking objects
    trackerlist = default_trackerlist(workdir, loggername, trackevery) if default_trackers else []
    if additional_trackers:
        trackerlist += additional_trackers
    
    if adageobj:
        ## funny behavior of multiprocessing Pools means that
        ## we can not have backendsubmit = multiprocsetup(2)    in the function sig
        ## so we only initialize them here
        if not backend:
            from backends import MultiProcBackend
            backend = MultiProcBackend(2)

        ## prep controller with backend
        controller = InMemoryController(adageobj, backend)

    run_polling_workflow(controller, coroutine, update_interval, trackerlist, maxsteps)




import json
import adage.adageobject
import adage.graph
import logging
log = logging.getLogger(__name__)

class DefaultAdageEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, adage.adageobject):
            return obj_to_json(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)

def noop_taskserializer(task):
    return 'unserializable_task'

def noop_proxyserializer(proxy):
    return 'unserializable_proxy'

def noop_ruleserializer(rule):
    return 'unserializable_rule'

def obj_to_json(adageobj, ruleserializer = noop_ruleserializer, taskserializer = noop_taskserializer, proxyserializer = noop_proxyserializer):
    dag, rules, applied = adageobj.dag, adageobj.rules, adageobj.applied_rules
    data = {'dag':None, 'rules':None, 'applied':None}

    data['dag'] = {'nodes':[], 'edges': []}
    for node in dag.nodes():
        nodeobj = dag.getNode(node)
        data['dag']['nodes']+=[node_to_json(nodeobj,taskserializer,proxyserializer)]

    data['dag']['edges'] += dag.edges()

    data['rules'] = []
    for rule in rules:
        data['rules'] += [ruleserializer(rule)]

    data['applied'] = []
    for rule in applied:
        data['applied'] += [ruleserializer(rule)]

    return data

def node_to_json(nodeobj,taskserializer,proxyserializer):
    # log.info('serializing node %s %s',nodeobj.name, proxyserializer(nodeobj.resultproxy))
    nodeinfo = {
        'id':nodeobj.identifier,
        'name':nodeobj.name,
        'task':taskserializer(nodeobj.task),
        'timestamps':{
            'defined': nodeobj.define_time,
            'submit': nodeobj.submit_time,
            'ready by': nodeobj.ready_by_time
        },
        'state':str(nodeobj.state),
        'proxy':proxyserializer(nodeobj.resultproxy)
    }
    return nodeinfo

def dag_from_json(dagdata,nodeclass,proxyclass,backend):
    dag = adage.graph.AdageDAG()

    for x in dagdata['nodes']:
        node = nodeclass.fromJSON(x)
        node.define_time = x['timestamps']['defined']
        node.submit_time = x['timestamps']['submit']
        node.ready_by_time = x['timestamps']['ready by']
        node.resultproxy = proxyclass.fromJSON(x['proxy']) if x['proxy'] else None
        node.backend = backend
        node.update_state()
        # log.info('deserialized node %s', node)
        dag.addNode(node)

    for x in dagdata['edges']:
        dag.addEdge(dag.getNode(x[0]),dag.getNode(x[1]))


    return dag



import adage.controllerutils as ctrlutils
import logging
import nodestate

log = logging.getLogger(__name__)

class BaseController(object):
    '''
    standard workflow controller, that keeps workflow state in memory at all times without any disk I/O or
    database access.
    '''

    def __init__(self, backend = None):
        '''
        :param backend: the desired backend to which to submit nodes
        '''
        self.backend  = backend

    def submit_nodes(self, nodes):
        '''
        :param nodes: a list of nodes to submit to the backend. 

        Nodes are expected to be provided in same form as what the controller returns.
        '''
        ctrlutils.submit_nodes(nodes, self.backend)

    def apply_rules(self, rules):
        '''
        :param rules: a list of rules to to apply to the workflow graph

        Rules are expected to be provided in same form as what the controller returns.
        '''
        ctrlutils.apply_rules(self.adageobj, rules)

    def applicable_rules(self):
        '''
        :return: return a list of rules whose predicate is fulfilled
        '''
        self.sync_backend()
        return ctrlutils.applicable_rules(self.adageobj)

    def submittable_nodes(self):
        '''
        :return: a list of nodes with sucessfull and completed upstream
        '''
        self.sync_backend()
        return ctrlutils.submittable_nodes(self.adageobj)

    def finished(self):
        '''
        :return: boolean indicating if nodes or rules are still left to be submitted/applied    
        '''
        self.sync_backend() #so that we are up to date
        return not ctrlutils.nodes_left_or_rule_applicable(self.adageobj)

    def successful(self):
        '''
        :return: boolean indicating workflow execution was successful
        '''
        self.sync_backend()
        failed = any(self.adageobj.dag.getNode(x).state == nodestate.FAILED for x in self.adageobj.dag.nodes())
        return (not failed)

    def validate(self):
        '''
        :return: validates internal execution order of workflow
        '''
        if not ctrlutils.validate_finished_dag(self.adageobj.dag):
            return False

        # if self.adageobj.rules:
        #     log.warning('some rules were not applied.')

        return True

    def sync_backend(self):
        '''
        :return: synchronize with backend to update workflow state
        '''
        return ctrlutils.sync_state(self.adageobj)

class InMemoryController(BaseController):
    def __init__(self, adageobj, backend):
        self.adageobj = adageobj
        super(InMemoryController, self).__init__(backend)



import nodestate
import time
import uuid
import logging

log = logging.getLogger(__name__)

class Node(object):
    def __init__(self,name,task,identifier = None, define_time = None):
        self.identifier = identifier or str(uuid.uuid4())
        self.name = name
        self.task = task

        # the timestamps
        self.define_time   = define_time or time.time()
        self.submit_time   = None
        self.ready_by_time = None

        # backend to update state against
        self.backend = None

        # relevant state data
        self.resultproxy = None
        self._result = None
        self._state = nodestate.DEFINED

    def __repr__(self):
        return '<Node name: {} id: {} state: {}>'.format(self.name,self.identifier,self.state)

    def update_state(self):
        #if we do not have a result object
        #that means it's not submitted yet
        if not self.resultproxy:
            self._state = nodestate.DEFINED
            return

        #if we have a resultobject
        #but the result is not ready
        #the node is still running
        if not self.backend.ready(self.resultproxy):
            self._state = nodestate.RUNNING
            return

        #if it's ready it's either successful
        #or failed
        if self.backend.successful(self.resultproxy):
            self._state  = nodestate.SUCCESS
            self._result = self.backend.result(self.resultproxy)
        else:
            self._state = nodestate.FAILED

        #it's ready so set time stamp it not already set
        if not self.ready_by_time:
            self.ready_by_time = time.time()
            log.info('node ready %s',self)

    @property
    def result(self):
        return self._result

    @property
    def state(self):
        return self._state

    def ready(self):
        return self.state in [nodestate.SUCCESS, nodestate.FAILED]

    def successful(self):
        return self.state == nodestate.SUCCESS



import networkx as nx
import logging
import dagstate
import time
import datetime

log = logging.getLogger(__name__)

def validate_finished_dag(dag):
    '''
    check for correct temporal execution order

    :param dag: graph object
    :return: ``True`` if order correct, ``False`` otherwise
    '''

    for node in dag:
        nodeobj = dag.getNode(node)
        if nodeobj.submit_time:
            for x in dag.predecessors(node):
                prednode = dag.getNode(x)
                if not nodeobj.submit_time > prednode.ready_by_time:
                    log.error('??? apparently {} was submitted before predesessor was ready: {}'.format(nodeobj, prednode))
                    log.error('node was submitted at: {} {}'.format(datetime.datetime.fromtimestamp(nodeobj.submit_time), nodeobj))
                    log.error('predecessor finished at: {} {}'.format(datetime.datetime.fromtimestamp(prednode.ready_by_time), prednode))
                    return False
    return True

def nodes_left_or_rule_applicable(adageobj):
    '''
    :param adageobj: the adage workflow object
 
    Main worklflow status function. It checks whether the overall workflow is finished by checking whether
    any eligible node (i.e. one without upstream failure) has not yet finished, or if any rules are still
    applicable.
    
    :param adageobj: the adage workflow object
    :return:
        - True any eligible node are waiting to be run, or still running or a rule is applicable
        - False in all other cases.
    '''

    dag,rules = adageobj.dag, adageobj.rules
    nodes_we_could_run = [node for node in dag.nodes() if not dagstate.upstream_failure(dag,dag.getNode(node))]
    nodes_running_or_defined = [x for x in nodes_we_could_run if dagstate.node_defined_or_running(dag.getNode(x))]

    if any(rule.applicable(adageobj) for rule in rules):
        return True

    log.debug('nodes we could run: %s',nodes_we_could_run)
    if nodes_running_or_defined:
        log.debug('%s nodes that could be run or are running are left.',len(nodes_running_or_defined))
        log.debug('nodes are: %s',[dag.node[n] for n in nodes_running_or_defined])
        return True

    if any(rule.applicable(adageobj) for rule in rules):
        return True


    log.info('no nodes can be run anymore and no rules are applicable')
    return False

def submittable_nodes(adageobj):
    '''
    main generator to iterate through nodes in the DAG and yield the ones that are submittable

    :param adageobj: the adage workflow object
    '''
    log.debug("process DAG")
    dag = adageobj.dag
    for node in nx.topological_sort(dag):
        nodeobj = dag.getNode(node)
        log.debug("working on node: %s with obj %s",node,nodeobj)
        if nodeobj.submit_time:
            log.debug("node already submitted. continue")
            continue;
        if dagstate.upstream_ok(dag,nodeobj):
            yield nodeobj
        # if dagstate.upstream_failure(dag,nodeobj):
        #     log.debug('not yielding node: %s due to upstream failure',node)

def submit_nodes(nodeobjs,backend):
    '''
    :param nodeobj: the node object, whose task to submit
    :param backend: the task execution backend
    :return: None

    basic submission of a task associated to a given node object
    '''

    for nodeobj in nodeobjs:
        # log.info('submitting node %s', nodeobj)
        nodeobj.resultproxy = backend.submit(nodeobj.task)
        nodeobj.submit_time = time.time()
        if not nodeobj.backend:
            nodeobj.backend = backend

def applicable_rules(adageobj):
    '''
    Generator to iterate over applicable rules.
    yields the applicable rules (and their indices in the rules array)

    :param adageobj: the adage workflow object
    '''
    for rule in adageobj.rules:
        if rule.applicable(adageobj):
            yield rule
        else:
            log.debug('rule %s not ready yet',rule)

def apply_rules(adageobj, rules):
    '''
    :param adageobj: the adage workflow object
    :param rules: list of rule to apply
    :return: None

    applies a number of rules. The order of application is an implementation detail
    and the client must not assume on any specific of application
    '''

    for rule in rules:
        # log.info('applying rule %s', rule)
        rule.apply(adageobj)
        rule_index = adageobj.rules.index(rule)
        log.debug('popping rule at index %s', rule_index)
        adageobj.applied_rules.append(adageobj.rules.pop(rule_index))


def sync_state(adageobj):
    '''
    Synchronize with Backend to update node processing status

    :param adageobj: the adage workflow object
    :return: None
    '''
    # log.info('sync state against backend')
    for node in adageobj.dag.nodes():
        #check node status one last time so we pick up the finishing times
        adageobj.dag.getNode(node).update_state()

def update_coroutine(adageobj):
    '''
    loops over applicable coroutines, applies them and manages the bookkeeping (moving rules from 'open' to 'applied')
    :param adageobj: the adage workflow object
    '''
    for i,rule in applicable_rules(adageobj):
        do_apply = yield rule
        if do_apply:
            log.debug('extending graph.')
            apply_rules(adageobj, [rule])
        yield



import logging

log = logging.getLogger(__name__)

def update_coroutine(controller):
    '''
    loops over applicable coroutines, applies them and manages the bookkeeping (moving rules from 'open' to 'applied')
    :param controller: the adage workflow controller
    '''
    for rule in controller.applicable_rules():
        do_apply = yield rule
        if do_apply:
            log.debug('extending graph.')
            controller.apply_rules([rule])
        yield

def update_dag(controller, decider):
    '''
    :param controller: the adage workflow controller
    :param decider: a decision coroutine. 

    Higher level DAG update routine that calls the basic update loop recursively (
    in order to apply as many DAG extensions as possible). The `decider` coroutine will
    receive a (rule, controller) tuple and is expected to return control by yielding
    a boolean value
    '''
    log.debug("update DAG by submitting nodes")
    anyapplied = False
    update_loop = update_coroutine(controller)
    for possible_rule in update_loop:
        log.debug('we could update this with rule: %s',possible_rule)
        command = decider.send((possible_rule,controller))
        if command:
            log.debug('we are in fact updating this..')
            anyapplied = True
        update_loop.send(command)
    #we changed the state so let's just recurse
    if anyapplied:
        log.debug('we applied a change, so we will recurse to see if we can apply anything else give updated state')
        update_dag(controller, decider)


def process_dag(controller,decider):
    '''
    main loop to go through nodes in the DAG and submit the onces that are submittable
    '''
    log.debug("process DAG by submitting nodes")
    for nodeobj in controller.submittable_nodes():
        do_submit = decider.send((nodeobj,controller))
        if do_submit:
            controller.submit_nodes([nodeobj])

def adage_coroutine(extend_decider,submit_decider):
    '''
    :param extend_decider: decision coroutine to decide whether to apply applicable rules
    :param submit_decider: decision coroutine to decide whether to submit applicable nodes

    the main loop as a coroutine, for sequential stepping (repeated polling) through the workflow.
    the loop will go through a update->submit->sync cycle while yielding control to the decision
    coroutines
    '''

    # after priming the coroutine, we yield right away until we get send a state object

    controller = yield

    # after receiving the state object, we yield and will start the loop once we regain controls
    yield

    #starting the loop
    while not controller.finished():
        update_dag(controller, extend_decider)
        process_dag(controller,submit_decider)
        # we're done for this tick, let others proceed
        yield controller
    log.info('exiting main polling coroutine')

def yes_man(messagestring = 'saying yes.'):
    '''trivial decision function that just returns True always'''
    # we yield until we receive some data via send()
    data = yield
    while True:
        log.debug(messagestring,data)
        #we yield True and wait again to receive some data
        value = True
        data = yield value

def setup_polling_execution(extend_decider = None, submit_decider = None):
    '''
    sets up the main couroutine and auxiliary decision coroutines for polling-style
    workflow exectuion. Optionally decision coroutines can be passed as parameters 
    (must already be primed)

    :param extend_decider: decision coroutine to decide whether to apply applicable rules
    :param submit_decider: decision coroutine to decide whether to submit applicable nodes
    '''

    if not extend_decider:
        extend_decider = yes_man('say yes to graph extension by %s')
        extend_decider.next() # prime

    if not submit_decider:
        submit_decider = yes_man('say yes to node submission of: %s')
        submit_decider.next() # prime

    ## prep main coroutine with deciders..
    log.info('preparing adage coroutine.')
    coroutine = adage_coroutine(extend_decider,submit_decider)
    coroutine.next() # prime the coroutine....

    return coroutine
